"""Simulator for Visible Detector"""
from copy import deepcopy

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pandorapsf as pp
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from pandorasat import PANDORASTYLE, VisibleDetector, get_logger

from . import __version__
from .docstrings import add_docstring
from .sim import Sim
from .utils import get_jitter

__all__ = ["VisibleSim"]

logger = get_logger("pandora-sim")


class VisibleSim(Sim):
    @add_docstring("ROI_size", "nROIs")
    def __init__(
        self,
        ROI_size=(50, 50),
        nROIs=9,
    ):
        """
        Visible Simulator for Pandora.
        """
        super().__init__(detector=VisibleDetector())
        self.ROI_size = ROI_size
        self.nROIs = nROIs

    @add_docstring("nROIs")
    def select_ROI_corners(self, nROIs, magnitude_limit=14):
        """Selects the corners of ROIs.

        This is currently a placeholder. SOC will provide direction on how ROIs will be selected.

        Parameters:
        -----------
        magnitude_limit : float
            Visual magnitude limit down to which ROI targets will be considered.
        """
        source_mag = deepcopy(np.asarray(self.source_catalog.mag))
        locations = deepcopy(self.locations)

        # This downweights sources far from the middle
        r = 1 - np.hypot(
            *(self.locations - np.asarray(self.detector.shape) / 2).T
        ) / np.hypot(*np.asarray(self.detector.shape))
        source_mag += -2.5 * np.log10(r)

        k = (source_mag < magnitude_limit) & (self.source_catalog.ruwe <= 1.2)
        locations, source_mag = locations[k], source_mag[k]
        size = np.asarray(self.ROI_size)
        crpix = self.wcs.wcs.crpix
        corners = [(-size[0] // 2 + crpix[0], -size[1] // 2 + crpix[1])]
        while len(corners) < nROIs:
            if len(locations) == 0:
                raise ValueError(f"Can not select {nROIs} ROIs")
            idx = np.argmin(source_mag)
            corner = np.round(locations[idx]).astype(int) - size // 2
            if ~np.any(
                [
                    (np.abs(c[0] - corner[0]) < size[0] // 2)
                    & (np.abs(c[1] - corner[1]) < size[1] // 2)
                    for c in corners
                ]
            ):
                corners.append(tuple(c for c in corner))

            k = ~np.in1d(np.arange(len(locations)), idx)
            locations = locations[k]
            source_mag = source_mag[k]
        return corners

    @add_docstring("ra", "dec", "theta")
    def point(self, ra, dec, roll):
        """
        Point the simulation in a direction.
        """
        super().point(ra=ra, dec=dec, roll=roll)
        logger.start_spinner("Building scene object...")
        self.scene = pp.Scene(
            self.locations - np.asarray(self.detector.shape) / 2,
            self.psf,
            self.detector.shape,
            corner=(-self.detector.shape[0] // 2, -self.detector.shape[1] // 2),
        )
        logger.stop_spinner()
        self.ROI_corners = self.select_ROI_corners(self.nROIs)

        logger.start_spinner("Building ROI scene object...")

        k = np.any(
            [
                (self.source_catalog.row > c[0])
                & (self.source_catalog.row < (c[0] + self.ROI_size[0]))
                & (self.source_catalog.column > c[1])
                & (self.source_catalog.column < (c[1] + self.ROI_size[1]))
                for c in self.ROI_corners
            ],
            axis=0,
        )

        self.roiscene = pp.ROIScene(
            self.locations[k] - np.asarray(self.detector.shape) / 2,
            self.psf,
            self.detector.shape,
            corner=(-self.detector.shape[0] // 2, -self.detector.shape[1] // 2),
            ROI_size=self.ROI_size,
            ROI_corners=self.ROI_corners,
            nROIs=self.nROIs,
        )
        logger.stop_spinner()

    def _get_source_catalog(self):
        source_catalog = super()._get_source_catalog(gbpmagnitude_range=(-6, 18))

        # we're assuming that Gaia B mag is very close to the Pandora visible magnitude
        vis_counts = np.zeros(len(source_catalog))
        vis_flux = np.zeros(len(source_catalog))
        wav = np.arange(100, 1000) * u.nm
        s = np.trapz(self.detector.sensitivity(wav), wav)
        for idx, m in enumerate(np.asarray(source_catalog.mag)):
            f = self.detector.flux_from_mag(m)
            vis_flux[idx] = f.value
            # counts in electron/u.second
            vis_counts[idx] = (f * s).to(u.electron / u.second).value

        source_catalog["counts"] = vis_counts
        source_catalog["flux"] = vis_flux
        return source_catalog

    @property
    def background_rate(self):
        return 4 * u.electron / u.second

    @add_docstring("nreads", "noise")
    def get_FFI(
        self,
        nreads: int = 50,
        noise=True,
    ):
        """Get a single frame of data as an FFI

        Returns:
        --------
        data : np.ndarray
            Returns a single FFI as a numpy array with dtype uint32.
        """
        int_time = self.detector.integration_time * nreads
        source_flux = (
            (np.asarray(self.source_catalog.counts) * u.electron / u.second) * int_time
        ).value.astype(int)

        # FFI has shape (nrows, ncolumns), in units of electrons.
        ffi = self.scene.model(source_flux)

        # Apply poisson (shot) noise, ffi now has shape  (nrows, ncolumns), units of electrons
        ffi = np.random.poisson(self.scene.model(source_flux)[0])
        if noise:
            # Apply background to every read, units of electrons
            ffi += np.random.poisson(
                (self.background_rate * int_time).value, size=ffi.shape
            ).astype(int)

            # # Apply a bias to every read which is a Gaussian with mean = bias * nreads value and std = (nreads * (read noise)**2)**0.5
            # We actually do this as a sum because otherwise the integer math doesn't work out...!?

            test_distribution = (
                np.random.normal(
                    loc=self.detector.bias.value,
                    scale=self.detector.read_noise.value,
                    size=(nreads, 10000),
                )
                .astype(int)
                .sum(axis=0)
            )
            ffi += np.random.normal(
                loc=test_distribution.mean(),
                scale=self.detector.read_noise.value * np.sqrt(nreads),
                size=(ffi.shape),
            ).astype(int)

            # Add poisson noise for the dark current to every frame, units of electrons
            ffi += np.random.poisson(
                lam=(self.detector.dark * int_time).value,
                size=ffi.shape,
            ).astype(int)

        # Apply gain
        #        ffi = self.detector.apply_gain(u.Quantity(ffi.ravel(), unit='electron')).value.reshape(ffi.shape)
        # Crap gain for now because gain calculations are wicked broken
        ffi *= 2

        # This is a bit hacky, but for FFIs we'll be ok. We do this because working with individual reads for FFI data is slow.
        ffi[ffi > (nreads * 2**16)] = nreads * 2**16
        return ffi

    @add_docstring(
        "nreads",
        "nframes",
        "start_time",
        "target_flux_function",
        "noise",
        "jitter",
        "output_type",
        "bin_frames",
    )
    def observe(
        self,
        nreads=50,
        nframes=100,
        start_time=Time("2000-01-01T12:00:00", scale="utc"),
        target_flux_function=None,
        noise=True,
        jitter=True,
        output_type="fits",
        bin_frames=10,
    ):
        """
        Returns an observation as though taken from the Pandora Observatory.

        Parameters:
        -----------

        Returns:
        --------
        result : np.ndarray or astropy.io.fits.HDUList
            Result of the simulation. Either a numpy array with shape (nROIs, nframes, nrows, ncolumns), or a fits format.

        """
        if not nreads / bin_frames == nreads // bin_frames:
            raise ValueError(
                "`bin_frames` must be a factor of `nreads` (e.g. `nreads=50` can be binned by a factor of 2, 5, 10... etc)"
            )

        nr = nreads // bin_frames
        integration_time = self.detector.integration_time * bin_frames
        dt = np.arange(nr * nframes) * integration_time.to(u.second).value
        time = start_time.jd + (dt * u.second).to(u.day).value

        k = np.any(
            [
                (self.source_catalog.row > c[0])
                & (self.source_catalog.row < (c[0] + self.ROI_size[0]))
                & (self.source_catalog.column > c[1])
                & (self.source_catalog.column < (c[1] + self.ROI_size[1]))
                for c in self.ROI_corners
            ],
            axis=0,
        )

        source_flux = (
            (np.asarray(self.source_catalog.counts[k]) * u.electron / u.second)
            * integration_time.to(u.second)
        ).value.astype(int)[:, None] * np.ones(nr * nframes)[None, :]

        if jitter:
            _, row, column, _ = get_jitter(
                rowstd=((1 * u.arcsecond) / self.detector.pixel_scale)
                .to(u.pixel)
                .value,
                colstd=((1 * u.arcsecond) / self.detector.pixel_scale)
                .to(u.pixel)
                .value,
                nframes=nframes * nr,
                frame_time=integration_time.to(u.second).value,
            )
            delta_pos = np.asarray([row, column])
        else:
            delta_pos = None
        if target_flux_function is not None:
            source_flux[0, :] *= target_flux_function(time)

        # Data has shape (nROIs, ntimes, nrows, ncolumns), units of electrons
        data = self.roiscene.model(source_flux, delta_pos)
        data[data < 0] = 0

        if noise:
            # Apply poisson (shot) noise, data now has shape  (nROIs, nreads, nrows, ncolumns), units of electrons
            data = np.random.poisson(data)

            # Apply background to every read, units of electrons
            data += np.random.poisson(
                (self.background_rate * integration_time.to(u.second)).value,
                size=data.shape,
            ).astype(int)

            # # Apply a bias to every read which is a Gaussian with mean = bias value and std = read noise
            # We have to estimate the mean bias when summed across reads because the integer math messes with the mean
            test_distribution = (
                np.random.normal(
                    loc=self.detector.bias.value,
                    scale=self.detector.read_noise.value,
                    size=(nreads, 10000),
                )
                .astype(int)
                .sum(axis=0)
            )

            data += np.random.normal(
                loc=test_distribution.mean() / nr,
                scale=self.detector.read_noise.value * np.sqrt(bin_frames),
                size=data.shape,
            ).astype(int)

            # Add poisson noise for the dark current to every frame, units of electrons
            data += np.random.poisson(
                lam=(self.detector.dark * integration_time.to(u.second)).value,
                size=data.shape,
            ).astype(int)

        # Data in units of DN
        #    data = self.detector.apply_gain(u.Quantity(data.ravel(), unit='electron')).value.reshape(data.shape)

        # Crap gain for now because gain calculations are wicked broken
        data *= 2
        # Any pixels greater than uint16 are maxed out (saturated)
        data[data > 2**16] = 2**16

        # bin down the data across the read dimension
        data = data.reshape((self.nROIs, nframes, nr, 50, 50)).sum(axis=2)

        if output_type == "array":
            return data
        elif output_type == "fits":
            return self._roi_to_fits(
                data,
                nreads=nreads,
                start_time=start_time,
            )

    def show_ROI(self):
        """Plot an example of an ROI."""
        d = self.observe(
            nreads=50, nframes=1, output_type="array", jitter=False, noise=True
        )[:, 0, :, :]
        vmin, vmax = np.percentile(d, 1), np.percentile(d, 90)
        # Find the next largest perfect square from the number of subarrays given
        next_square = int(np.ceil(np.sqrt(self.nROIs)) ** 2)
        n = int(np.sqrt(next_square))

        _, ax = plt.subplots(n, n, figsize=(7, 6), sharex=True, sharey=True)
        plt.subplots_adjust(hspace=0.03, wspace=0.05)
        for idx in range(self.nROIs):
            l = np.where(np.arange(self.nROIs).reshape((n, n)) == idx)
            with plt.style.context(PANDORASTYLE):
                im = ax[l[0][0], l[1][0]].imshow(
                    d[idx],
                    vmin=vmin,
                    vmax=vmax,
                    interpolation="nearest",
                    origin="lower",
                    cmap="Greys",
                )
                if l[1][0] == 0:
                    ax[l[0][0], l[1][0]].set(ylabel="Row\n[subarray pixel]")
                if l[0][0] == (n - 1):
                    ax[l[0][0], l[1][0]].set(xlabel="Column\n[subarray pixel]")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label(
            f"Counts after {(self.detector.integration_time * 50).to_string(format='latex')} [DN]"
        )
        plt.suptitle(
            f"{self.detector.name}\n(RA:{self.ra:.02f}, Dec:{self.dec:.02f}, Roll:{self.roll:.02f})"
        )
        return ax

    def show_FFI(self, ax=None):
        """Plot an example of an FFI."""
        ffi = self.get_FFI(nreads=50)

        if ax is None:
            _, ax = plt.subplots(figsize=(7, 6))

        with plt.style.context(PANDORASTYLE):
            im = ax.imshow(
                ffi,
                vmin=np.percentile(ffi, 10),
                vmax=np.percentile(ffi, 99),
                interpolation="nearest",
                origin="lower",
                cmap="Greys",
            )
            for corner in self.ROI_corners:
                r, c = (
                    np.asarray([0, 0, self.ROI_size[0], self.ROI_size[0], 0])
                    + corner[0],
                    np.asarray([0, self.ROI_size[1], self.ROI_size[1], 0, 0])
                    + corner[1],
                )
                ax.plot(c, r, c="r")
            cbar = plt.colorbar(im, ax=ax)
            cbar.set_label(
                f"Counts after {(self.detector.integration_time * 50).to_string(format='latex')} [DN]"
            )
            ax.set(
                xlabel="Column [pixel]",
                ylabel="Row [pixel]",
                title=f"{self.detector.name}\n(RA:{self.ra:.02f}, Dec:{self.dec:.02f}, Roll:{self.roll:.02f})",
            )
        return ax

    @property
    def primary_kwds(self):
        return {
            "EXTNAME": ("PRIMARY", "name of extension"),
            "NEXTEND": (2, "number of standard extensions"),
            "SIMDATA": (True, "simulated data"),
            "SCIDATA": (False, "science data"),
            "TELESCOP": ("NASA Pandora", "telescope"),
            "INSTRMNT": ("VISDA", "instrument"),
            "CREATOR": ("Pandora DPC", "creator of this product"),
            "CRSOFTV": ("v" + str(__version__), "creator software version"),
            "TARG_RA": (self.ra.value, "target right ascension [deg]"),
            "TARG_DEC": (self.dec.value, "target declination [deg]"),
            "EXPDELAY": (-1, "exposure time delay [ms]"),
            "RICEX": (-1, "bit noise parameter for Rice compression"),
            "RICEY": (-1, "bit noise parameter for Rice compression"),
        }

    @property
    def roi_kwds(self):
        return {
            "NAXIS": (2, "number of array dimensions"),
            "NAXIS1": (
                len(self.ROI_corners[0]),
                "length of dimension 1",
            ),
            "NAXIS2": (len(self.ROI_corners), "length of dimension 2"),
            "PCOUNT": (0, "number of group parameters"),
            "GCOUNT": (1, "number of groups"),
            "TFIELDS": (2, "number of table fields"),
            "TTYPE1": ("Column", "table field 1 type"),
            "TFORM1": ("I21", "table field 1 format"),
            "TUNIT1": ("pix", "table field 1 unit"),
            "TBCOL1": (1, ""),
            "TTYPE2": ("Row", "table field 2 type"),
            "TFORM2": ("I21", "table field 2 format"),
            "TUNIT2": ("pix", "table field 2 unit"),
            "TBCOL2": (22, ""),
            "EXTNAME": ("ROITABLE", "name of extension"),
            "NROI": (
                len(self.ROI_corners),
                "number of regions of interest",
            ),
            "ROISTRTX": (
                -1,
                "region of interest origin position in column",
            ),
            "ROISTRTY": (-1, "region of interest origin position in row"),
            "ROISIZEX": (-1, "region of interest size in column"),
            "ROISIZEY": (-1, "region of interest size in row"),
        }

    def _roi_to_fits(self, data, start_time, nreads=50, integration_time=None):
        if integration_time is None:
            integration_time = self.detector.integration_time
        n_arrs, nframes, nrows, ncols = data.shape

        corstime = int(
            np.floor((start_time - Time("2000-01-01T12:00:00", scale="utc")).sec)
        )
        finetime = int(corstime % 1 * 10**9 // 1)

        primary_kwds = self.primary_kwds
        primary_kwds["FRMSREQD"] = (nframes, "number of frames requested")
        primary_kwds["FRMSCLCT"] = (nframes, "number of frames collected")
        primary_kwds["NUMCOAD"] = (nreads, "number of frames coadded")
        primary_kwds["FRMTIME"] = (
            (nreads * integration_time).to(u.second).value,
            "time in each frame [s]",
        )
        primary_kwds["CORSTIME"] = (
            corstime,
            "seconds since the TAI Epoch (12PM Jan 1, 2000)",
        )
        primary_kwds["FINETIME"] = (finetime, "nanoseconds added to CORSTIME seconds")
        primary_hdu = fits.PrimaryHDU()
        for key, value in primary_kwds.items():
            primary_hdu.header[key] = value

        # Find the next largest perfect square from the number of subarrays given
        next_square = int(np.ceil(np.sqrt(n_arrs)) ** 2)
        sq_sides = int(np.sqrt(next_square))

        padding = np.zeros((next_square - n_arrs, nframes, nrows, ncols), dtype=int)
        subarrays = np.append(data, padding, axis=0)
        subarrays = subarrays.reshape((sq_sides, sq_sides, *data.shape[1:])).transpose(
            [2, 0, 1, 3, 4]
        )
        image_data = np.block(
            [[subarrays[:, i, j] for j in range(sq_sides)] for i in range(sq_sides)]
        ).astype(data.dtype)
        image_kwds = {
            "NAXIS": (3, "number of array dimensions"),
            "NAXIS1": (
                image_data.shape[1],
                "first axis size",
            ),  # might need to change these
            "NAXIS2": (image_data.shape[2], "second axis size"),
            "NAXIS3": (image_data.shape[0], "third axis size"),
            "EXTNAME": ("SCIENCE", "extension name"),
            "TTYPE1": ("COUNTS", "data title: raw pixel counts"),
            "TFORM1": ("J", "data format: images of unsigned 32-bit integers"),
            "TUNIT1": ("count", "data units: count"),
        }
        image_hdu = fits.ImageHDU(image_data)
        for key, value in image_kwds.items():
            image_hdu.header[key] = value

        for card in self.wcs.to_header(relax=True).cards:
            image_hdu.header.append(card)

        table_hdu = fits.table_to_hdu(Table(self.ROI_corners))
        for key, value in self.roi_kwds.items():
            table_hdu.header[key] = value

        hdulist = fits.HDUList([primary_hdu, image_hdu, table_hdu])
        return hdulist



import functools
from typing import Callable

from astropy.time import Time

PARAMETERS_DOCSTRINGS = {
    "SC_Resets1": (
        int,
        "Number of reset frames at the start of the first integration of exposure",
    ),
    "SC_Resets2": (
        int,
        "Number of resent frames at the start of 1 through n integrations of exposure",
    ),
    "SC_DropFrames1": (
        int,
        "Number of dropped frames after reset of any integration of exposure",
    ),
    "SC_DropFrames2": (
        int,
        "Number of dropped frames in every group of integrations of exposure except the last group",
    ),
    "SC_DropFrames3": (
        int,
        "Number of dropped frames in the last group of each integration of exposure",
    ),
    "SC_ReadFrames": (
        int,
        "Number of frames read during each group of integration of exposure",
    ),
    "SC_Groups": (int, "Number of groups per integration of exposure"),
    "SC_Integrations": (int, "Number of integrations per exposure"),
    "ra": (float, "Right Ascension of the pointing"),
    "dec": (float, "Declination of the pointing"),
    "theta": (float, "Roll angle of the pointing"),
    "ROI_size": (tuple, "Size in pixels of each region of interest"),
    "nROIs": (int, "Number of regions of interest"),
    "nreads": (
        int,
        "Number of reads to co-add together to make each frame. Default for Pandora VISDA is 50.",
    ),
    "nframes": (int, "Number of frames to return. Default is 100 frames."),
    "start_time": (
        Time,
        "Time to start the observation. This is only used if returning a fits file, or supplying a flux function.",
    ),
    "target_flux_function": (
        (Callable, None),
        "Function governing the generation of the time series from the subarray. Default is None.",
    ),
    "noise": (
        bool,
        "Flag determining whether noise is simulated and included. Default is True.",
    ),
    "jitter": (
        bool,
        "Flag determining whether jitter is simulated and included. Default is True.",
    ),
    "output_type": (
        str,
        'String flag to determine the result output. Valid strings are "fits" or "array"',
    ),
    "bin_frames": (
        int,
        "If `nreads` is high, many reads must be calculated per frame stored. To reduce this, set `bin_frames`. If `bin_frames=10`, only `nreads/10` reads will be calculated, each with an exposure time of `self.detector.integration_time * bin_frames`.",
    ),
}


# Decorator to add common parameters to docstring
def add_docstring(*param_names):
    def decorator(func):
        param_docstring = ""
        if func.__doc__:
            # Determine the current indentation level
            lines = func.__doc__.splitlines()
            if len(lines[0]) == 0:
                indent = len(lines[1]) - len(lines[1].lstrip())
            else:
                indent = len(lines[0]) - len(lines[0].lstrip())
        else:
            indent = 0
        indent_str = " " * indent
        for name in param_names:
            if name in PARAMETERS_DOCSTRINGS:
                dtype, desc = PARAMETERS_DOCSTRINGS[name]
                if isinstance(dtype, tuple):
                    dtype_str = " or ".join(
                        [t._name if hasattr(t, "_name") else t.__name__][0]
                        for t in dtype
                        if t is not None
                    )
                    dtype_str += " or None" if None in dtype else ""
                else:
                    dtype_str = dtype.__name__
                param_docstring += (
                    f"{indent_str}{name} : {dtype_str}\n{indent_str}    {desc}\n"
                )
        existing_docstring = func.__doc__ or ""
        if "Parameters" in existing_docstring:
            func.__doc__ = (
                existing_docstring.split("---\n")[0]
                + "---\n"
                + param_docstring
                + "---\n".join(existing_docstring.split("---\n")[1:])
            )
        else:
            func.__doc__ = (
                existing_docstring
                + f"\n{indent_str}Parameters:\n{indent_str}-----------\n"
                + param_docstring
            )
        return func

    return decorator


# Decorator to inherit docstring from base class
def inherit_docstring(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    if func.__doc__ is None:
        for base in func.__qualname__.split(".")[0].__bases__:
            base_func = getattr(base, func.__name__, None)
            if base_func and base_func.__doc__:
                func.__doc__ = base_func.__doc__
                break
    return wrapper



# Standard library
import warnings
from copy import copy
from typing import Dict, Optional

# Third-party
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.convolution import Gaussian1DKernel, convolve
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time
from astroquery import log as asqlog
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive

asqlog.setLevel("ERROR")
# Third-party

FRAME_BIT_DICT = {"reset": 1, "read": 2, "drop": 4}


def get_planets(coord: SkyCoord, radius: u.Quantity = 20 * u.arcsecond) -> dict:
    """Largish default radius for high proper motion targets this breaks
    Returns a dictionary of dictionaries with planet parameters.
    """
    try:
        coord2000 = coord.apply_space_motion(Time(2016, format="jyear"))
    except ValueError:
        coord2000 = coord
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        planets_tab = NasaExoplanetArchive.query_region(
            table="pscomppars", coordinates=coord2000, radius=radius
        )
        if len(planets_tab) != 0:
            attrs = ["pl_orbper", "pl_tranmid", "pl_trandur", "pl_trandep"]
            planets = {
                letter: {
                    attr: planets_tab[planets_tab["pl_letter"] == letter][attr][
                        0
                    ].unmasked
                    for attr in attrs
                }
                for letter in planets_tab["pl_letter"]
            }
            # There's an error in the NASA exoplanet archive units that makes duration "days" instead of "hours"
            for planet in planets:
                planets[planet]["pl_trandur"] = (
                    planets[planet]["pl_trandur"].value * u.hour
                )
        else:
            planets = {}
    return planets


def get_jitter(
    rowstd: float = 1,
    colstd: float = 0.3,
    thetastd: float = 0.0005,
    correlation_time: float = 1 * u.second,
    nframes: int = 50,
    frame_time: float = 0.2 * u.second,
    seed=None,
):
    """Returns some random, time correlated jitter time-series.

    Parameters:
    ----------
    rowstd: float
        Standard deviation of jitter in pixels in row/y axis
    colstd: float
        Standard deviation of jitter in pixels in col/x axis
    thetastd: float
        Standard deviation of jitter in degrees in y axis
    correlation_time: float
        The timescale over which data is correlated in seconds.
        Increase this value for smoother time-series
    nframes: int
        Number of frames to generate
    frame_time: float
        The time spacing for each frame
    seed: Optional, int
        Optional seed for random walk

    Returns:
    --------
    time : np.ndarray
        Time array in seconds
    row: np.ndarray
        Jitter in the row/y axis in pixels
    col: np.ndarray
        Jitter in the column/x axis in pixels
    theta: np.ndarray
        Jitter in angle in degrees
    """
    time = np.arange(nframes) * frame_time  # noqa:F811
    tstd = (correlation_time / frame_time).value

    def jitter_func(std):
        f = np.random.normal(0, std, size=nframes)
        return convolve(f, Gaussian1DKernel(tstd)) * tstd**0.5

    jitter = []
    for idx, std, unit in zip(
        [0, 1, 2], [rowstd, colstd, thetastd], [u.pixel, u.pixel, u.deg]
    ):
        if seed is not None:
            np.random.seed(seed + idx)
        jitter.append(jitter_func(std) * unit)

    return time, *jitter


def get_integrations(
    SC_Resets1,
    SC_Resets2,
    SC_DropFrames1,
    SC_DropFrames2,
    SC_DropFrames3,
    SC_ReadFrames,
    SC_Groups,
    SC_Integrations,
):
    """
    Function to generate the integration frames as a list of arrays containing flags denoting
    the status of each frame. The frames are grouped by status and ordered chronologically.

    Parameters
    ----------
    SC_Resets1 : int
        Number of reset frames at the start of the first integration of exposure
    SC_Resets2 : int
        Number of resent frames at the start of 1 through n integrations of exposure
    SC_DropFrames1 : int
        Number of dropped frames after reset of any integration of exposure
    SC_DropFrames2 : int
        Number of dropped frames in every group of integrations of exposure except the last group
    SC_DropFrames3 : int
        Number of dropped frames in the last group of each integration of exposure
    SC_ReadFrames : int
        Number of frames read during each group of integration of exposure
    SC_Groups : int
        Number of groups per integration of exposure
    SC_Integrations : int
        Number of integrations per exposure

    Returns
    -------
    integrations : list
        List of arrays containing each frame in the integrations grouped in chronological order
        with each integer element representing that frame's status.
    """
    cintn = [
        np.zeros(SC_Resets2, int) + FRAME_BIT_DICT["reset"],
        *[
            np.zeros(SC_ReadFrames, int) + FRAME_BIT_DICT["read"],
            np.zeros(SC_DropFrames2, int) + FRAME_BIT_DICT["drop"],
        ]
        * (SC_Groups - 1),
        np.zeros(SC_ReadFrames, int) + FRAME_BIT_DICT["read"],
    ]
    cint1 = copy(cintn)
    cint1.pop(0)

    cint1.insert(0, np.zeros(SC_Resets1, int) + FRAME_BIT_DICT["reset"])
    cint1.insert(1, np.zeros(SC_DropFrames1, int) + FRAME_BIT_DICT["drop"])

    integrations = [cint1]
    for idx in np.arange(SC_Integrations - 1):
        integrations.append(copy(cintn))

    integrations[-1].append(np.zeros(SC_DropFrames3, int) + FRAME_BIT_DICT["drop"])
    return integrations


def get_plot_vectors(inte):
    """
    Helper function to provide plotting information for each frame to the `plot_integrations`
    function.

    Parameters
    ----------
    inte : array
        The array of integer flags representing each frame in a single integration of exposure

    Returns
    -------
    vectors : list of arrays
        List of arrays specifying the plotting parameters for each individual frame
    """
    s = np.where(inte | FRAME_BIT_DICT["reset"] == FRAME_BIT_DICT["reset"])[0]
    if len(s) > 0:
        s = s[-1]
        y = np.hstack(
            [
                np.zeros(s + 1, dtype=int),
                np.arange(s + 1, len(inte), dtype=int) - s + 1,
            ]
        )
    else:
        y = np.arange(0, len(inte), dtype=int)

    cols = np.vstack(
        [
            inte | FRAME_BIT_DICT["reset"] == FRAME_BIT_DICT["reset"],
            inte | FRAME_BIT_DICT["read"] == FRAME_BIT_DICT["read"],
            inte | FRAME_BIT_DICT["drop"] == FRAME_BIT_DICT["drop"],
        ]
    ).T
    return np.hstack([y[:, None], cols])


def plot_nirda_integrations(
    SC_Resets1,
    SC_Resets2,
    SC_DropFrames1,
    SC_DropFrames2,
    SC_DropFrames3,
    SC_ReadFrames,
    SC_Groups,
    SC_Integrations,
):
    """
    Function to plot the integrations of the NIRDA visually given values describing
    the integration strategy.

    Parameters
    ----------
    SC_Resets1 : int
        Number of reset frames at the start of the first integration of exposure
    SC_Resets2 : int
        Number of resent frames at the start of 1 through n integrations of exposure
    SC_DropFrames1 : int
        Number of dropped frames after reset of any integration of exposure
    SC_DropFrames2 : int
        Number of dropped frames in every group of integrations of exposure except the last group
    SC_DropFrames3 : int
        Number of dropped frames in the last group of each integration of exposure
    SC_ReadFrames : int
        Number of frames read during each group of integration of exposure
    SC_Groups : int
        Number of groups per integration of exposure
    SC_Integrations : int
        Number of integrations per exposure
    """
    integrations = get_integrations(
        SC_Resets1,
        SC_Resets2,
        SC_DropFrames1,
        SC_DropFrames2,
        SC_DropFrames3,
        SC_ReadFrames,
        SC_Groups,
        SC_Integrations,
    )
    cadences = [
        np.sum([(i == FRAME_BIT_DICT["read"]).all() for i in inte if len(i) > 0])
        for inte in integrations
    ]
    integrations = [np.hstack(idx) for idx in integrations]
    dat = np.vstack([get_plot_vectors(inte) for inte in integrations])
    with plt.style.context("seaborn-v0_8-white"):
        fig, ax = plt.subplots(figsize=(np.max([5, len(dat) // 25]), 2))
        ax.plot(dat[:, 0], ls="--", c="grey")
        ax.scatter(
            np.arange(len(dat))[dat[:, 1].astype(bool)],
            dat[:, 0][dat[:, 1].astype(bool)],
            c="k",
            label="Reset",
        )
        ax.scatter(
            np.arange(len(dat))[dat[:, 2].astype(bool)],
            dat[:, 0][dat[:, 2].astype(bool)],
            c="lightblue",
            label="Frames",
        )
        ax.scatter(
            np.arange(len(dat))[dat[:, 3].astype(bool)],
            dat[:, 0][dat[:, 3].astype(bool)],
            c="r",
            label="Dropped Frames",
            marker="x",
        )
        ax.set(
            xlabel="Time",
            ylabel="Counts",
            xticks=[],
            yticks=[],
            title=(
                f"{SC_Integrations} Integrations, "
                + f"{SC_Groups} Groups\n{SC_ReadFrames} Read/group, "
                + f"{SC_DropFrames2} Drop/group\n{len(dat)} total reads, "
                + f"{np.sum(cadences)} total cadences"
            ),
        )
        ax.legend(frameon=True, bbox_to_anchor=(1.1, 1.05))
    return fig


def save_to_FITS(
    data: np.ndarray,
    filename: str,
    primary_kwds: Dict,
    image_kwds: Dict,
    roitable: bool = False,
    roitable_kwds: Optional[Dict] = None,
    roi_data: Optional[np.ndarray] = None,
    overwrite: bool = True,
):
    primary_hdu = fits.PrimaryHDU()

    for key, value in primary_kwds.items():
        primary_hdu.header[key] = value

    image_hdu = fits.ImageHDU(data)

    for key, value in image_kwds.items():
        image_hdu.header[key] = value

    hdu_list = [primary_hdu, image_hdu]

    if roitable:
        # table_hdu = fits.TableHDU(roi_data)
        table_hdu = fits.table_to_hdu(roi_data)

        for key, value in roitable_kwds.items():
            table_hdu.header[key] = value

        hdu_list.append(table_hdu)

    hdul = fits.HDUList(hdu_list)

    hdul.writeto(filename, overwrite=overwrite)



"""Simulator for Visible Detector"""
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pandorapsf as pp
from astropy.io import fits
from astropy.time import Time
from pandorasat import PANDORASTYLE, NIRDetector, get_logger
from pandorasat.phoenix import get_phoenix_model

from . import __version__
from .docstrings import add_docstring
from .sim import Sim
from .utils import FRAME_BIT_DICT, get_integrations, get_jitter

__all__ = ["NIRSim"]

logger = get_logger("pandora-sim")


class NIRSim(Sim):
    def __init__(
        self,
    ):
        """
        NIR Simulator for Pandora.
        """
        super().__init__(detector=NIRDetector())
        self.psf = self.psf.freeze_dimension(row=0 * u.pixel, column=0 * u.pixel)

    @add_docstring("ra", "dec", "theta")
    def point(self, ra, dec, roll):
        """
        Point the simulation in a direction.
        """
        super().point(ra=ra, dec=dec, roll=roll)
        logger.start_spinner("Building trace scene object...")
        self.tracescene = pp.TraceScene(
            self.locations,
            psf=self.psf,
            shape=self.detector.subarray_size,
            corner=(0, 0),
            wav_bin=1,
        )
        logger.stop_spinner()

    def _get_source_catalog(self):
        source_catalog = super()._get_source_catalog(gbpmagnitude_range=(-6, 21))

        logger.start_spinner("Interpolating PHOENIX spectra...")
        spectra = np.zeros((len(source_catalog), self.psf.trace_wavelength.shape[0]))
        for idx, teff, logg, j in zip(
            range(len(source_catalog)),
            np.nan_to_num(np.asarray(source_catalog.teff), 5777),
            np.nan_to_num(np.asarray(source_catalog.logg), 4.5),
            source_catalog.jmag,
        ):
            if teff < 2000:
                logger.warn(
                    f"{teff}K source in catalog is not modelable. Setting to 2000K."
                )
                teff = 2000
            if teff > 10000:
                logger.warn(
                    f"{teff}K source in catalog is not modelable. Setting to 10000K."
                )
                teff = 10000
            wav, spec = get_phoenix_model(teff=teff, logg=logg, jmag=j)
            spectra[idx, :] = self.psf.integrate_spectrum(wav, spec)
        logger.stop_spinner()

        # Units of electrons/s
        self.spectra = spectra * u.electron / u.s
        return source_catalog

    @property
    def background_rate(self):
        return 4 * u.electron / u.second

    @add_docstring(
        "SC_Resets1",
        "SC_Resets2",
        "SC_DropFrames1",
        "SC_DropFrames2",
        "SC_DropFrames3",
        "SC_ReadFrames",
        "SC_Groups",
        "SC_Integrations",
        "start_time",
        "noise",
        "jitter",
        "output_type",
    )
    def observe(
        self,
        SC_Resets1=1,
        SC_Resets2=1,
        SC_DropFrames1=0,
        SC_DropFrames2=16,
        SC_DropFrames3=0,
        SC_ReadFrames=4,
        SC_Groups=2,
        SC_Integrations=10,
        start_time=Time("2000-01-01T12:00:00", scale="utc"),
        target_spectrum_function=None,
        noise=True,
        jitter=True,
        output_type="fits",
    ):
        """
        Returns an observation as though taken from the Pandora Observatory.

        Parameters:
        -----------

        Returns:
        --------
        result : np.ndarray or astropy.io.fits.HDUList
            Result of the simulation. Either a numpy array with shape (nROIs, nframes, nrows, ncolumns), or a fits format.

        """

        integration_time = self.detector.frame_time()
        integration_info = get_integrations(
            SC_Resets1=SC_Resets1,
            SC_Resets2=SC_Resets2,
            SC_DropFrames1=SC_DropFrames1,
            SC_DropFrames2=SC_DropFrames2,
            SC_DropFrames3=SC_DropFrames3,
            SC_ReadFrames=SC_ReadFrames,
            SC_Groups=SC_Groups,
            SC_Integrations=SC_Integrations,
        )
        cadences = np.sum(
            [
                np.sum(
                    [(i == FRAME_BIT_DICT["read"]).all() for i in inte if len(i) > 0]
                )
                for inte in integration_info
            ]
        )
        integration_arrays = [np.hstack(idx) for idx in integration_info]

        nreads = np.sum([len(i) for i in integration_arrays])
        # one where a frame is being reset.
        resets = np.hstack(integration_arrays) != 1

        dt = np.arange(nreads) * integration_time.to(u.second).value
        time = start_time.jd + (dt * u.second).to(u.day).value

        source_flux = (
            self.spectra.T[:, :, None]
            * np.ones(nreads)[None, None, :]
            * integration_time
            * resets.astype(float)
        )

        if jitter:
            _, row, column, _ = get_jitter(
                rowstd=((1 * u.arcsecond) / self.detector.pixel_scale)
                .to(u.pixel)
                .value,
                colstd=((1 * u.arcsecond) / self.detector.pixel_scale)
                .to(u.pixel)
                .value,
                nframes=nreads,
                frame_time=integration_time.to(u.second).value,
            )
            delta_pos = np.asarray([row, column])
        else:
            delta_pos = None
        if target_spectrum_function is not None:
            source_flux[0, :, :] *= target_spectrum_function(time)

        # Data has shape (ntargets, nwavelength, ntimes), units of electrons
        data = self.tracescene.model(source_flux, delta_pos)
        data[data < 0] = 0

        if noise:
            # Apply poisson (shot) noise, data now has shape  (ntargets, nwavelength, ntimes), units of electrons
            data = np.random.poisson(data)

            # Apply background to every read, units of electrons
            data += np.random.poisson(
                (self.background_rate * integration_time.to(u.second)).value,
                size=data.shape,
            ).astype(int) * resets[:, None, None].astype(int)

            # Read Noise
            data += np.random.normal(
                loc=0,
                scale=self.detector.read_noise.value,
                size=data.shape,
            ).astype(int)

            # Add poisson noise for the dark current to every frame, units of electrons
            data += np.random.poisson(
                lam=(self.detector.dark * integration_time.to(u.second)).value,
                size=data.shape,
            ).astype(int)

        # Data in units of DN
        #    data = self.detector.apply_gain(u.Quantity(data.ravel(), unit='electron')).value.reshape(data.shape)

        # Crap gain for now because gain calculations are wicked broken
        data = (data.astype(float) * 0.5).astype(int)
        bias = self.detector.bias.value * 0.5
        bias_std = self.detector.bias_uncertainty.value * 0.5
        # Any pixels greater than uint16 are maxed out (saturated)
        data[(data + bias) > 2**16] = 2**16

        # Splits data into arrays representing each integration
        data_by_integration = np.array_split(
            data, np.cumsum(np.asarray([i.shape for i in integration_arrays])[:, 0])
        )[:-1]

        def get_group_masks(integration_info):
            """Function to create a boolean mask which is true for all the frames in each group."""
            masks = []
            for idx, i in enumerate(integration_info):
                if ((i == 2).all()) & (len(i) > 0):
                    masks.append(
                        (
                            np.hstack(
                                [
                                    np.zeros(len(i), bool)
                                    if idx != jdx
                                    else np.ones(len(i), bool)
                                    for jdx, i in enumerate(integration_info)
                                ]
                            )
                        )
                    )
            return np.asarray(masks)

        result = []
        for cdx, d, info in zip(range(cadences), data_by_integration, integration_info):
            # Cumulative sum for reading up the ramp
            d = np.cumsum(d, axis=0).astype(np.uint16)
            if noise:
                d += np.random.normal(loc=bias, scale=bias_std, size=d.shape).astype(
                    np.uint16
                )

            # Masks for each group
            group_masks = get_group_masks(info)
            # Output is the average in each group across the time dimension, after being cast in uint32 arrays.
            result.append(
                np.asarray(
                    [d[mask].astype(np.uint32).mean(axis=0) for mask in group_masks],
                    dtype=np.uint32,
                )
            )

        result = np.asarray(result, dtype=np.uint32)
        if output_type == "array":
            return result
        if output_type == "fits":
            return self._array_to_fits(
                result,
                start_time=start_time,
                SC_Resets1=SC_Resets1,
                SC_Resets2=SC_Resets2,
                SC_DropFrames1=SC_DropFrames1,
                SC_DropFrames2=SC_DropFrames2,
                SC_DropFrames3=SC_DropFrames3,
                SC_ReadFrames=SC_ReadFrames,
                SC_Groups=SC_Groups,
                SC_Integrations=SC_Integrations,
            )
        else:
            raise ValueError(f"Can not parse output type `{output_type}`")

    def show_subarray(self, **kwargs):
        """Plot an example subarray observation."""
        data = self.observe(
            SC_Integrations=1, jitter=False, noise=True, output_type="array"
        )[0, :, :, :]
        d = data.astype(int)[-1] - data.astype(int)[0]
        vmin, vmax = kwargs.pop("vmin", np.percentile(d, 1)), kwargs.pop(
            "vmax", -np.percentile(d, 1)
        )
        cmap = kwargs.pop("cmap", "Greys")
        _, ax = plt.subplots(figsize=(3, 6))
        with plt.style.context(PANDORASTYLE):
            im = ax.imshow(
                d,
                vmin=vmin,
                vmax=vmax,
                interpolation="nearest",
                origin="lower",
                cmap=cmap,
            )
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label("Counts In (Last Group - First Group) [DN]")
        ax.set(
            title=f"{self.detector.name}\n(RA:{self.ra:.02f}, Dec:{self.dec:.02f}, Roll:{self.roll:.02f})",
            xlabel="Column [pixel]",
            ylabel="Row [pixel]",
        )
        return ax

    @property
    def primary_kwds(self):
        return {
            "EXTNAME": ("PRIMARY", "name of extension"),
            "NEXTEND": (1, "number of standard extensions"),
            "SIMDATA": (True, "simulated data"),
            "SCIDATA": (False, "science data"),
            "TELESCOP": ("NASA Pandora", "telescope"),
            "INSTRMNT": ("NIRDA", "instrument"),
            "CREATOR": ("Pandora DPC", "creator of this product"),
            "CRSOFTV": ("v" + str(__version__), "creator software version"),
            "TARG_RA": (self.ra.value, "target right ascension [deg]"),
            "TARG_DEC": (self.dec.value, "target declination [deg]"),
            "EXPDELAY": (-1, "exposure time delay [ms]"),
            "RICEX": (-1, "bit noise parameter for Rice compression"),
            "RICEY": (-1, "bit noise parameter for Rice compression"),
        }

    def _array_to_fits(
        self,
        data,
        start_time,
        SC_Resets1,
        SC_Resets2,
        SC_DropFrames1,
        SC_DropFrames2,
        SC_DropFrames3,
        SC_ReadFrames,
        SC_Groups,
        SC_Integrations,
    ):

        integration_time = self.detector.frame_time()

        primary_kwds = self.primary_kwds
        primary_kwds["SC_Resets1"] = (
            SC_Resets1,
            "Number of reset frames at the start of the first integration of exposure",
        )
        primary_kwds["SC_Resets2"] = (
            SC_Resets2,
            "Number of resent frames at the start of 1 through n integrations of exposure",
        )
        primary_kwds["SC_DropFrames1"] = (
            SC_DropFrames1,
            "Number of dropped frames after reset of any integration of exposure",
        )
        primary_kwds["SC_DropFrames2"] = (
            SC_DropFrames2,
            "Number of dropped frames in every group of integrations of exposure except the last group",
        )
        primary_kwds["SC_DropFrames3"] = (
            SC_DropFrames3,
            "Number of dropped frames in the last group of each integration of exposure",
        )
        primary_kwds["SC_ReadFrames"] = (
            SC_ReadFrames,
            "Number of frames read during each group of integration of exposure",
        )
        primary_kwds["SC_Groups"] = (
            SC_Groups,
            "Number of groups per integration of exposure",
        )
        primary_kwds["SC_Integrations"] = (
            SC_Integrations,
            "Number of integrations per exposure",
        )

        corstime = int(
            np.floor((start_time - Time("2000-01-01T12:00:00", scale="utc")).sec)
        )
        finetime = int(corstime % 1 * 10**9 // 1)

        primary_kwds["FRAMTIME"] = (
            self.detector.frame_time().to(u.second).value,
            "Frame time of each read",
        )
        primary_kwds["GRPTIME"] = (
            SC_ReadFrames * integration_time.to(u.second).value,
            "Exposure time of each group",
        )
        primary_kwds["CORSTIME"] = (
            corstime,
            "seconds since the TAI Epoch (12PM Jan 1, 2000)",
        )
        primary_kwds["FINETIME"] = (finetime, "nanoseconds added to CORSTIME seconds")
        primary_hdu = fits.PrimaryHDU()
        for key, value in primary_kwds.items():
            primary_hdu.header[key] = value

        image_hdu = fits.ImageHDU(data)
        image_kwds = {
            "NAXIS1": (
                data.shape[3],
                "pixel columns",
            ),  # might need to change these
            "NAXIS2": (data.shape[2], "pixel rows"),
            "NAXIS3": (data.shape[1], "number of groups"),
            "NAXIS4": (data.shape[0], "number of integrations"),
            "EXTNAME": ("SCIENCE", "extension name"),
            "TTYPE1": ("COUNTS", "data title: raw pixel counts"),
            "TFORM1": ("J", "data format: images of unsigned 32-bit integers"),
            "TUNIT1": ("count", "data units: count"),
        }

        for key, value in image_kwds.items():
            image_hdu.header[key] = value
        for card in self.wcs.to_header(relax=True).cards:
            image_hdu.header.append(card)

        hdulist = fits.HDUList([primary_hdu, image_hdu])

        return hdulist



"""Abstract base class for a Simulator object"""
from abc import ABC, abstractmethod
from copy import deepcopy

import astropy.units as u
import numpy as np
import pandas as pd
import pandorapsf as pp
import pandorasat as ps
from pandorasat import get_logger

from .docstrings import add_docstring

__all__ = ["Sim"]

logger = get_logger("pandora-sim")


class Sim(ABC):
    def __repr__(self):
        if hasattr(self, "ra"):
            return f"{self.detector.name} Simulation [({self.ra:.3f}, {self.dec:.3f})]"
        else:
            return f"{self.detector.name} Simulation [no pointing information]"

    def __init__(self, detector):
        self.detector = detector
        logger.start_spinner("Loading PSF..")
        self.psf = pp.PSF.from_name(self.detector.name)
        logger.stop_spinner()

    @add_docstring("ra", "dec", "theta")
    @abstractmethod
    def point(self, ra: u.Quantity, dec: u.Quantity, roll: u.Quantity):
        self.ra, self.dec, self.roll = ra, dec, roll
        self.wcs = self.detector.get_wcs(self.ra, self.dec, theta=self.roll)

        logger.start_spinner("Finding nearby sources...")
        self.source_catalog = self._get_source_catalog()
        self.locations = np.asarray(
            [self.source_catalog.row, self.source_catalog.column]
        ).T
        logger.stop_spinner()

    def world_to_pixel(self, ra, dec, distortion=True):
        """Helper function. This function ensures we keep the row-major convention in pandora-sim.

        Parameters:
        -----------
        ra : float
            Right Ascension to be converted to pixel position.
        dec : float
            Declination to be converted to pixel position.
        distortion : bool
            Flag whether to account for the distortion in the WCS when converting from RA/Dec
            to pixel position. Default is True.

        Returns
        -------
        np.ndarray
            Row and column positions of each provided RA and Dec.
        """
        coords = np.vstack(
            [
                ra.to(u.deg).value if isinstance(ra, u.Quantity) else ra,
                dec.to(u.deg).value if isinstance(dec, u.Quantity) else dec,
            ]
        ).T
        if distortion:
            column, row = self.wcs.all_world2pix(coords, 0).T
        else:
            column, row = self.wcs.wcs_world2pix(coords, 0).T
        return np.vstack([row, column])

    def pixel_to_world(self, row, column, distortion=True):
        """Helper function. This function ensures we keep the row-major convention in pandora-sim.

        Parameters:
        -----------
        row : float
            Pixel row position to be converted to sky coordinates.
        column : float
            Pixel column position to be converted to sky coordinates.
        distortion : bool
            Flag whether to account for the distortion in the WCS when converting from pixel position
            to sky coordinates. Default is True.

        Returns
        -------
        np.ndarray
            RA and Dec of input pixel positions.
        """
        coords = np.vstack(
            [
                column.to(u.pixel).value if isinstance(column, u.Quantity) else column,
                row.to(u.pixel).value if isinstance(row, u.Quantity) else row,
            ]
        ).T
        if distortion:
            return self.wcs.all_pix2world(coords, 0).T * u.deg
        else:
            return self.wcs.wcs_pix2world(coords, 0).T * u.deg

    def _get_source_catalog(self, distortion: bool = True, **kwargs) -> pd.DataFrame:
        """Gets the source catalog of an input target

        Parameters
        ----------
        target_name : str
            Target name to obtain catalog for.
        distortion : bool
            Whether to apply a distortion to the WCS when mapping RA and Dec of catalog targets
            to detector pixels. Default is True.
        **kwargs
            Additional arguments passed to the ps.utils.get_sky_catalog function.

        Returns
        -------
        sky_catalog: pd.DataFrame
            Pandas dataframe of all the sources near the target
        """

        # This is fixed for visda, for now
        if hasattr(self.detector, "subarray_size"):
            shape = self.detector.subarray_size
        else:
            shape = self.detector.shape
        radius = np.hypot(*np.asarray(shape) // 2)
        radius = ((radius * u.pixel) * self.detector.pixel_scale).to(u.deg).value

        # Get location and magnitude data
        cat = ps.utils.get_sky_catalog(
            self.ra, self.dec, radius=radius * u.deg, **kwargs
        )
        ra, dec, mag = cat["coords"].ra.deg, cat["coords"].dec.deg, cat["bmag"]
        pix_coords = self.world_to_pixel(ra, dec, distortion=distortion)

        k = (
            np.abs(pix_coords[0] - shape[0] / 2)
            < (shape[0] / 2 + self.psf_shape[0] / 2)
        ) & (
            np.abs(pix_coords[1] - shape[1] / 2)
            < (shape[1] / 2 + self.psf_shape[1] / 2)
        )

        new_cat = deepcopy(cat)
        for key, item in new_cat.items():
            new_cat[key] = item[k]
        pix_coords, ra, dec, mag = (
            pix_coords[:, k],
            ra[k],
            dec[k],
            mag[k],
        )
        source_catalog = (
            pd.DataFrame(
                np.vstack(
                    [
                        ra,
                        dec,
                        mag,
                        *pix_coords,
                        new_cat["jmag"],
                        new_cat["teff"].value,
                        new_cat["logg"],
                        new_cat["RUWE"],
                        new_cat["ang_sep"].value,
                    ]
                ).T,
                columns=[
                    "ra",
                    "dec",
                    "mag",
                    "row",
                    "column",
                    "jmag",
                    "teff",
                    "logg",
                    "ruwe",
                    "ang_sep",
                ],
            )
            .drop_duplicates(["ra", "dec", "mag"])
            .reset_index(drop=True)
        )
        return source_catalog

    @property
    def psf_shape(self):
        p = self.detector.trace_pixel.value
        p = p.max() - p.min()
        return (p + 6, 6)



__version__ = "1.0.4"
# Standard library
import os  # noqa

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))
TESTDIR = "/".join(PACKAGEDIR.split("/")[:-2]) + "/tests/"

# Standard library
import logging  # noqa: E402

logging.basicConfig()
logger = logging.getLogger("pandorasim")


from .nirsim import NIRSim  # noqa
from .utils import plot_nirda_integrations  # noqa
from .visiblesim import VisibleSim  # noqa
